@@ -67,7 +67,7 @@
  *
  * mod_python.c 
  *
- * $Id: mod_python.c,v 1.8 2000/05/22 12:14:39 grisha Exp $
+ * $Id: mod_python.c,v 1.12 2000/05/29 16:55:11 grisha Exp $
  *
  * See accompanying documentation and source code comments 
  * for details.
@@ -93,9 +93,9 @@
                         Declarations
  ******************************************************************/
 
-#define VERSION_COMPONENT "mod_python/2.0"
+#define VERSION_COMPONENT "mod_python/2.1"
 #define MODULENAME "mod_python.apache"
-#define INITSTRING "mod_python.apache.init()"
+#define INITFUNC "init"
 #define INTERP_ATTR "__interpreter__"
 
 /* Are we in single interpreter mode? */
@@ -105,20 +105,11 @@ static int single_mode = 0;
  * (In a Python dictionary) */
 static PyObject * interpreters = NULL;
 
-/* The CallBack object. This variable is used as
- * a way to pass a pointer between  SetCallBack()
- * function and make_obcallback(). Nothing else. 
- * Don't rely on its value.                      */
-static PyObject *obCallBack = NULL;
-
-/* This function is used with ap_register_cleanup() */
-void noop(void *data) {};
-void python_decref(void *object);
-
 /* some forward declarations */
-PyObject * make_obcallback(const char *module, const char *initstring);
+void python_decref(void *object);
+PyObject * make_obcallback();
 PyObject * tuple_from_array_header(const array_header *ah);
-PyObject * get_obcallback(const char *name, server_rec * req);
+PyObject * get_obcallback(const char *name, request_rec * req);
 
 /*********************************
            Python things 
@@ -128,13 +119,11 @@ PyObject * get_obcallback(const char *name, server_rec * req);
  *********************************/
 
 /* froward declarations */
-static PyObject * SetCallBack(PyObject *self, PyObject *args);
 static PyObject * log_error(PyObject *self, PyObject *args);
 static PyObject * make_table(PyObject *self, PyObject *args);
 
 /* methods of _apache */
 static struct PyMethodDef _apache_module_methods[] = {
-  {"SetCallBack",               (PyCFunction)SetCallBack,      METH_VARARGS},
   {"log_error",                 (PyCFunction)log_error,        METH_VARARGS},
   {"make_table",                (PyCFunction)make_table,       METH_VARARGS},
   {NULL, NULL} /* sentinel */
@@ -150,10 +139,9 @@ typedef struct tableobject {
     pool            *pool;
 } tableobject;
 
-static PyTypeObject tableobjecttype;
-
-#define is_tableobject(op) ((op)->ob_type == &tableobjecttype)
-
+static void table_dealloc(tableobject *self);
+static PyObject * table_getattr(PyObject *self, char *name);
+static PyObject * table_repr(tableobject *self);
 static PyObject * tablegetitem  (tableobject *self, PyObject *key );
 static PyObject * table_has_key (tableobject *self, PyObject *args );
 static PyObject * table_keys    (tableobject *self);
@@ -162,18 +150,38 @@ static int tablesetitem         (tableobject *self, PyObject *key, PyObject *val
 static int tb_setitem           (tableobject *self, const char *key, const char *val);
 static tableobject * make_tableobject(table * t);
 
-static PyMethodDef tablemethods[] = {
-  {     "keys",                 (PyCFunction)table_keys,    1},
-  {     "has_key",              (PyCFunction)table_has_key, 1},
-  { NULL, NULL } /* sentinel */
-};
-
 static PyMappingMethods table_mapping = {
     (inquiry)       tablelength,           /*mp_length*/
     (binaryfunc)    tablegetitem,          /*mp_subscript*/
     (objobjargproc) tablesetitem,          /*mp_ass_subscript*/
 };
 
+static PyTypeObject tableobjecttype = {
+    PyObject_HEAD_INIT(&PyType_Type)
+    0,
+    "mptable",
+    sizeof(tableobject),
+    0,
+    (destructor) table_dealloc,     /*tp_dealloc*/
+    0,                              /*tp_print*/
+    (getattrfunc) table_getattr,    /*tp_getattr*/
+    0,                              /*tp_setattr*/
+    0,                              /*tp_compare*/
+    (reprfunc) table_repr,          /*tp_repr*/
+    0,                              /*tp_as_number*/
+    0,                              /*tp_as_sequence*/
+    &table_mapping,                 /*tp_as_mapping*/
+    0,                              /*tp_hash*/
+};
+
+#define is_tableobject(op) ((op)->ob_type == &tableobjecttype)
+
+static PyMethodDef tablemethods[] = {
+  {     "keys",                 (PyCFunction)table_keys,    1},
+  {     "has_key",              (PyCFunction)table_has_key, 1},
+  { NULL, NULL } /* sentinel */
+};
+
 /* another forward */
 tableobject * headers_in(request_rec *req);
 
@@ -187,7 +195,26 @@ typedef struct serverobject {
     PyObject       *next;
 } serverobject;
 
-static PyTypeObject serverobjecttype;
+static void server_dealloc(serverobject *self);
+static PyObject * server_getattr(serverobject *self, char *name);
+
+static PyTypeObject serverobjecttype = {
+    PyObject_HEAD_INIT(&PyType_Type)
+    0,
+    "server",
+    sizeof(serverobject),
+    0,
+    (destructor) server_dealloc,     /*tp_dealloc*/
+    0,                               /*tp_print*/
+    (getattrfunc) server_getattr,    /*tp_getattr*/
+    0,                               /*tp_setattr*/
+    0,                               /*tp_compare*/
+    0,                               /*tp_repr*/
+    0,                               /*tp_as_number*/
+    0,                               /*tp_as_sequence*/
+    0,                               /*tp_as_mapping*/
+    0,                               /*tp_hash*/
+};
 
 #define is_serverobject(op) ((op)->ob_type == &serverobjecttype)
 
@@ -229,7 +256,26 @@ typedef struct connobject {
   PyObject     *base_server;
 } connobject;
 
-static PyTypeObject connobjecttype;
+static void conn_dealloc(connobject *self);
+static PyObject * conn_getattr(connobject *self, char *name);
+
+static PyTypeObject connobjecttype = {
+    PyObject_HEAD_INIT(&PyType_Type)
+    0,
+    "conn",
+    sizeof(connobject),
+    0,
+    (destructor) conn_dealloc,       /*tp_dealloc*/
+    0,                               /*tp_print*/
+    (getattrfunc) conn_getattr,      /*tp_getattr*/
+    0,                               /*tp_setattr*/
+    0,                               /*tp_compare*/
+    0,                               /*tp_repr*/
+    0,                               /*tp_as_number*/
+    0,                               /*tp_as_sequence*/
+    0,                               /*tp_as_mapping*/
+    0,                               /*tp_hash*/
+};
 
 #define is_connobject(op) ((op)->ob_type == &connobjecttype)
 
@@ -272,10 +318,10 @@ typedef struct requestobject {
     int              header_sent;
 } requestobject;
 
-static PyTypeObject requestobjecttype;
-
-#define is_requestobject(op) ((op)->ob_type == &requestobjecttype)
 
+static void request_dealloc(requestobject *self);
+static PyObject * request_getattr(requestobject *self, char *name);
+static int request_setattr(requestobject *self, char *name, PyObject *value);
 static PyObject * req_send_http_header     (requestobject *self, PyObject *args);
 static PyObject * req_get_basic_auth_pw    (requestobject *self, PyObject *args);
 static PyObject * req_write                (requestobject *self, PyObject *args);
@@ -285,6 +331,26 @@ static PyObject * req_get_options          (requestobject *self, PyObject *args)
 static PyObject * req_get_dirs             (requestobject *self, PyObject *args);
 static PyObject * req_add_common_vars      (requestobject *self, PyObject *args);
 
+static PyTypeObject requestobjecttype = {
+    PyObject_HEAD_INIT(&PyType_Type)
+    0,
+    "request",
+    sizeof(requestobject),
+    0,
+    (destructor) request_dealloc,    /*tp_dealloc*/
+    0,                               /*tp_print*/
+    (getattrfunc) request_getattr,   /*tp_getattr*/
+    (setattrfunc) request_setattr,   /*tp_setattr*/
+    0,                               /*tp_compare*/
+    0,                               /*tp_repr*/
+    0,                               /*tp_as_number*/
+    0,                               /*tp_as_sequence*/
+    0,                               /*tp_as_mapping*/
+    0,                               /*tp_hash*/
+};
+
+#define is_requestobject(op) ((op)->ob_type == &requestobjecttype)
+
 static PyMethodDef requestobjectmethods[] = {
     {"send_http_header",     (PyCFunction) req_send_http_header,     METH_VARARGS},
     {"get_basic_auth_pw",    (PyCFunction) req_get_basic_auth_pw,    METH_VARARGS},
@@ -670,7 +736,7 @@ static PyObject * table_keys(tableobject *self)
 /**
  ** copy_table
  **
- *   Merge two tables into one. Matching ley values 
+ *   Merge two tables into one. Matching key values 
  *   in second overlay the first.
  */
 
@@ -690,51 +756,6 @@ static void copy_table(table *t1, table *t2)
 	    ap_table_set(t1, elts[i].key, elts[i].val);
 }
 
-/**
- ** print_table
- **
- *   Print apache table. Only used for debugging.
- */
-
-static void print_table(table * t)
-{
-    array_header *ah;
-    table_entry *elts;
-    int i;
-
-    ah = ap_table_elts (t);
-    elts = (table_entry *) ah->elts;
-    i = ah->nelts;
-
-    while (i--)
-	if (elts[i].key)
-	    printf("   %s: \t%s\n", elts[i].key, elts[i].val);
-}
-
-/**
- ** print_array
- **
- *   Print apache array (of strings). Only used for debugging.
- */
-
-static void *print_array(array_header *ah)
-{
-    int i;
-    char **elts;
-
-    elts = (char **)ah->elts;
-
-    for (i = 0; i < ah->nelts; ++i) {
-	printf("%s ", elts[i]);
-    }
-    printf("\n");
-
-    return NULL;
-    
-    /* avoid compiler warning */
-    print_array(ah);
-
-}
 
 /********************************
   ***  end of table object ***
@@ -975,7 +996,8 @@ static requestobject * make_requestobject(request_rec *req)
     result->header_sent = 0;
 
     _Py_NewReference(result);
-    ap_register_cleanup(req->pool, (PyObject *)result, python_decref, noop);
+    ap_register_cleanup(req->pool, (PyObject *)result, python_decref, 
+			ap_null_cleanup);
 
     return result;
 }
@@ -1350,76 +1372,6 @@ static PyObject * req_add_common_vars(requestobject *self, PyObject *args)
   ***  end of request object ***
  ********************************/
 
-/**
- ** table2dict()
- **
- *      Converts Apache table to Python dictionary.
- */
-
-static PyObject * table2dict(table * t)
-{
-
-    PyObject *result, *val;
-    array_header *ah;
-    table_entry *elts;
-    int i;
-
-    result = PyDict_New();
-
-    ah = ap_table_elts (t);
-    elts = (table_entry *) ah->elts;
-    i = ah->nelts;
-
-    while (i--)
-	if (elts[i].key)
-	{
-	    val = PyString_FromString(elts[i].val);
-	    PyDict_SetItemString(result, elts[i].key, val);
-	    Py_DECREF(val);
-	}
-
-    return result;
-
-    /* prevent compiler warning about unused 
-       function, this statement never reached */
-    table2dict(NULL);
-
-}
-
-/**
- ** python_dict_merge
- ** 
- *   Merges two Python dictionaries into one, overlaying
- *   items in the first argument by the second. Returns new
- *   reference.
- */
-
-void python_dict_merge(PyObject *d1, PyObject *d2)
-{
-
-    /*
-     * there are faster ways to do this if we were to bypas
-     * the Python API and use functions from dictobject.c
-     * directly
-     */
-    
-    PyObject *items;
-    PyObject *keyval;
-    PyObject *x;
-    PyObject *y;
-    int i;
-
-    items = PyDict_Items(d2);
-    for (i = 0; i < PyList_Size(items); i++) {
-	keyval = PyList_GetItem(items, i);
-	x = PyTuple_GetItem(keyval, 0);
-	y = PyTuple_GetItem(keyval, 1);
-	PyDict_SetItem(d1, PyTuple_GetItem(keyval, 0),
-		       PyTuple_GetItem(keyval, 1));
-    }
-    Py_DECREF(items);
-}
-
 /******************************************************************
    ***        end of Python objects and their methods        ***
  ******************************************************************/
@@ -1456,87 +1408,6 @@ void python_init(server_rec *s, pool *p)
     PyThreadState *tstate = NULL;
     PyObject *x;
 
-    /* initialize serverobjecttype */
-    PyTypeObject sot = {
-	PyObject_HEAD_INIT(&PyType_Type)
-	0,
-	"server",
-	sizeof(serverobject),
-	0,
-	(destructor) server_dealloc,     /*tp_dealloc*/
-	0,                               /*tp_print*/
-	(getattrfunc) server_getattr,    /*tp_getattr*/
-	0,                               /*tp_setattr*/
-	0,                               /*tp_compare*/
-	0,                               /*tp_repr*/
-	0,                               /*tp_as_number*/
-	0,                               /*tp_as_sequence*/
-	0,                               /*tp_as_mapping*/
-	0,                               /*tp_hash*/
-    };
-
-    /* initialize connobjecttype */
-    PyTypeObject cot = {
-	PyObject_HEAD_INIT(&PyType_Type)
-	0,
-	"conn",
-	sizeof(connobject),
-	0,
-	(destructor) conn_dealloc,       /*tp_dealloc*/
-	0,                               /*tp_print*/
-	(getattrfunc) conn_getattr,      /*tp_getattr*/
-	0,                               /*tp_setattr*/
-	0,                               /*tp_compare*/
-	0,                               /*tp_repr*/
-	0,                               /*tp_as_number*/
-	0,                               /*tp_as_sequence*/
-	0,                               /*tp_as_mapping*/
-	0,                               /*tp_hash*/
-    };
-
-    /* initialize requestobjecttype */
-    PyTypeObject rot = {
-	PyObject_HEAD_INIT(&PyType_Type)
-	0,
-	"request",
-	sizeof(requestobject),
-	0,
-	(destructor) request_dealloc,    /*tp_dealloc*/
-	0,                               /*tp_print*/
-	(getattrfunc) request_getattr,   /*tp_getattr*/
-	(setattrfunc) request_setattr,   /*tp_setattr*/
-	0,                               /*tp_compare*/
-	0,                               /*tp_repr*/
-	0,                               /*tp_as_number*/
-	0,                               /*tp_as_sequence*/
-	0,                               /*tp_as_mapping*/
-	0,                               /*tp_hash*/
-    };
-
-    /* initialize tableobjecttype */
-    PyTypeObject tt = {
-	PyObject_HEAD_INIT(&PyType_Type)
-	0,
-	"mptable",
-	sizeof(tableobject),
-	0,
-	(destructor) table_dealloc,     /*tp_dealloc*/
-	0,                              /*tp_print*/
-	(getattrfunc) table_getattr,    /*tp_getattr*/
-	0,                              /*tp_setattr*/
-	0,                              /*tp_compare*/
-	(reprfunc) table_repr,          /*tp_repr*/
-	0,                              /*tp_as_number*/
-	0,                              /*tp_as_sequence*/
-	&table_mapping,                 /*tp_as_mapping*/
-	0,                              /*tp_hash*/
-    };
-
-    serverobjecttype = sot;
-    connobjecttype = cot;
-    requestobjecttype = rot;
-    tableobjecttype = tt;
-
     /* mod_python version */
     ap_add_version_component(VERSION_COMPONENT);
 
@@ -1550,7 +1421,6 @@ void python_init(server_rec *s, pool *p)
 
 	/* initialze the interpreter */
 	Py_Initialize();
-               
 
 #ifdef WITH_THREAD
 	/* create and acquire the interpreter lock */
@@ -1559,42 +1429,21 @@ void python_init(server_rec *s, pool *p)
 	/* create the obCallBack dictionary */
 	interpreters = PyDict_New();
 
-	if (! interpreters)
-	{
+	if (! interpreters) {
 	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, s,
 			 "python_init: PyDict_New() failed! No more memory?");
 	    exit(1);
 	}
-	  
-	/*  Initialize _apache. 
-	 *  This makes an _apache module available for import, but remember,
-	 *  the user should use "import apache" not "_apache". "_apache" 
-	 *  is for internal use only. (The only time it's needed is to assign 
-	 *  obCallBack)
-	 */
-
-	/* make obCallBack */
-	obcallback = make_obcallback(NULL, NULL);
-
-	/* get the current thread state */
-	tstate = PyThreadState_Get();
-
-	/* cast PyThreadState * to long and save it in obCallBack */
-	x = PyInt_FromLong((long) tstate);
-	PyObject_SetAttrString(obcallback, INTERP_ATTR, x);
-	Py_DECREF(x);
-
-	/* save the obCallBack */
-	PyDict_SetItemString(interpreters, "global_interpreter", obcallback);
 
+	/* make obCallBack for the global interpeter */
+	obcallback = get_obcallback(NULL, NULL);
+	
 #ifdef WITH_THREAD
-
+	
 	/* release the lock; now other threads can run */
 	PyEval_ReleaseLock();
 #endif
-
     }
-
 }
 
 /**
@@ -1706,125 +1555,63 @@ static void *python_merge_dir_config(pool *p, void *cc, void *nc)
 static const char *python_directive(cmd_parms *cmd, void * mconfig, 
 				    char *key, const char *val)
 {
-
-    int directs_offset = XtOffsetOf(py_dir_config, directives);
-    int dirs_offset = XtOffsetOf(py_dir_config, dirs);
-    int config_dir_offset = XtOffsetOf(py_dir_config, config_dir);
-    char *s;
-    table *directives;
-    table *dirs;
-
-    directives = *(table **)(mconfig + directs_offset);
-    dirs = *(table **)(mconfig + dirs_offset);
-
-    ap_table_set(directives, ap_pstrdup(cmd->pool, key), 
-		 ap_pstrdup(cmd->pool, val));
+    py_dir_config *conf;
+    
+    conf = (py_dir_config *) mconfig;
+    ap_table_set(conf->directives, key, val);
 
     /* remember the directory where the directive was found */
-    s = *(char **)(mconfig + config_dir_offset);
-    if (s) {
-	ap_table_set(dirs, ap_pstrdup(cmd->pool, key), s);
+    if (conf->config_dir) {
+	ap_table_set(conf->dirs, key, conf->config_dir);
     }
     else {
-	ap_table_set(dirs, ap_pstrdup(cmd->pool, key), "");
+	ap_table_set(conf->dirs, key, "");
     }
 
     return NULL;
+
 }
 
-
-
-
 /**
  ** make_obcallback
  **
  *      This function instantiates an obCallBack object. 
  *      NOTE: The obCallBack object is instantiated by Python
- *      code. This C module calls into Python code which sets the 
- *      reference by calling SetCallBack function back into C from
- *      Python. Thus the name "CallBack". 
- * 
- * Here is a visual:
- *
- * - Begin C code
- * .   PyRun_SimpleString("init()")
- * =   Begin Python code
- * .     obCallBack = CallBack()
- * .     import _apache # this module is in C
- * .     _apache.SetCallBack(obCallBack)
- * --    Begin C code
- * .       // assigns to global obCallBack
- * --    End of C code
- * =   End of Python code
- * - we're back in C, but now global obCallBack has a value
+ *      code. This C module calls into Python code which returns 
+ *      the reference to obCallBack.
  */
 
-PyObject * make_obcallback(const char *module, const char *initstring)
+PyObject * make_obcallback()
 {
 
-  char buff[256];
+    PyObject *m;
+    PyObject *obCallBack;
+    char buff[256];
 
-  /*  In the old verion of this module these varuables were assigned by
-   *  PythonInitFunction directive, in the new version they'll be
-   *  NULL and get assigned here.
-   */
+    /* This makes _apache appear imported, and subsequent
+     * >>> import _apache 
+     * will not give an error.
+     */
+    Py_InitModule("_apache", _apache_module_methods);
 
-  if (! module)
-    module = MODULENAME;
+    /* Now execute the equivalent of
+     * >>> import <module>
+     * >>> <initstring>
+     * in the __main__ module to start up Python.
+     */
 
-  if (! initstring)
-    initstring = INITSTRING;
-
-  /* This makes _apache appear imported, and subsequent
-   * >>> import _apache 
-   * will not give an error.
-   */
-  Py_InitModule("_apache", _apache_module_methods);
-
-  /* Now execute the equivalent of
-   * >>> import sys
-   * >>> import <module>
-   * >>> <initstring>
-   * in the __main__ module to start up Python.
-   */
-
-  sprintf(buff, "import %s\n", module);
-
-  if (PyRun_SimpleString(buff))
-    {
-      fprintf(stderr, "PythonInitFunction: could not import %s.\n", module);
-      exit(1);
+    if (! ((m = PyImport_ImportModule(MODULENAME)))) {
+	fprintf(stderr, "make_obcallback(): could not import %s.\n", MODULENAME);
+	exit(1);
     }
-
-  sprintf(buff, "%s\n", initstring);
-
-  if (PyRun_SimpleString(buff))
-    {
-      fprintf(stderr, "PythonInitFunction: could not call %s.\n",
-	       initstring);
-      exit(1);
+    
+    if (! ((obCallBack = PyObject_CallMethod(m, INITFUNC, NULL)))) {
+	fprintf(stderr, "make_obcallback(): could not call %s.\n",
+		INITFUNC);
+	exit(1);
     }
-
-  /* we can't see it, but the execution has moved into a Python script
-   * which is doing something like this:
-   *
-   * >>> import _apache
-   * >>> _apache.SetCallBack(someCallBackObject)
-   *
-   * By now (i.e while you're reading this comment), 
-   * SetCallBack() has been called and obCallBack is assigned. 
-   */
-
-  if (! obCallBack) 
-    {
-      fprintf(stderr, "PythonInitFunction: after %s no callback object found.\n", 
-	       initstring);
-      exit(1);
-    }
-
-  /* Wow, this worked! */
-
-  return obCallBack;
+    
+    return obCallBack;
 
 }
 
@@ -1846,7 +1633,7 @@ PyObject * make_obcallback(const char *module, const char *initstring)
  *      name NULL means use global interpreter.
  */
 
-PyObject * get_obcallback(const char *name, server_rec * server)
+PyObject * get_obcallback(const char *name, request_rec * req)
 {
 
   PyObject * obcallback = NULL;
@@ -1884,22 +1671,21 @@ PyObject * get_obcallback(const char *name, server_rec * server)
       if (! tstate) {
 
 	  /* couldn't create an interpreter, this is bad */
-	  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, server,
+	  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req,
 		       "get_obcallback: Py_NewInterpreter() returned NULL. No more memory?");
 	  return NULL;
       }
       
       /* create an obCallBack */
-      obcallback = make_obcallback(NULL, NULL);
+      obcallback = make_obcallback();
 
-      /* cast PyThreadState * to long and save it in obCallBack */
-      p = PyInt_FromLong((long) tstate);
+      /* cast PyThreadState and save it in obCallBack  as CObject */
+      p = PyCObject_FromVoidPtr((void *) tstate, NULL);
       PyObject_SetAttrString(obcallback, INTERP_ATTR, p);
       Py_DECREF(p);
 
       /* save the obCallBack */
       PyDict_SetItemString(interpreters, (char *)name, obcallback);
-      
     }
 
   return  obcallback;
@@ -1947,96 +1733,61 @@ static PyObject * log_error(PyObject *self, PyObject *args)
     return Py_None;
 }
 
-/**
- ** SetCallBack - assign a CallBack object
- **
- *      This function must be called from Python upon executing
- *      the initstring parameter, like this
- *
- *       >>> import _apache
- *       >>> _apache.SetCallBack(instance)
- *
- *      to provide the instance as the CallBack object.
- */
-
-static PyObject * SetCallBack(PyObject *self, PyObject *args)
-{
-
-    PyObject *callback;
-
-    /* see if CallBack is in *args */
-    if (! PyArg_ParseTuple(args, "O", &callback)) 
-	return NULL;
-
-    /* store the object, incref */
-    obCallBack = callback;  
-    Py_INCREF(obCallBack);
-
-    /* return None */
-    Py_INCREF(Py_None);
-    return Py_None;
-}
-
 /**
  ** get_request_object
  **
- *      This creates or retrieves a previously created request object 
+ *      This creates or retrieves a previously created request object.
+ *      The pointer to request object is stored in tstate->dict, a 
+ *      special dictionary provided by Python for per-thread state.
  */
 
-static void get_request_object(request_rec *req, requestobject **request_obj)
-
+static requestobject *get_request_object(request_rec *req, 
+					 PyThreadState *tstate)
 {
-
+    requestobject *request_obj;
     char *s;
     char s2[40];
 
-    /* see if there is a request object out there already */
-
+    /* see if there is a request object already */
     /* XXX there must be a cleaner way to do this, atoi is slow? */
     /* since tables only understand strings, we need to do some conversion */
-
+    
     s = (char *) ap_table_get(req->notes, "python_request_ptr");
     if (s) {
-	*request_obj = (requestobject *) atoi(s);
-	/* we have one reference at creation that is shared, no INCREF here*/
-	return;
+	request_obj = (requestobject *) atoi(s);
+	return request_obj;
     }
     else {
-	/* ap_add_cgi_vars() is necessary for path-translated for example */
-	/* ap_add_cgi_vars() sometimes recurses, so we temporarily release the lock */
-
-	/* XXX - why do I need this? Still true as of ver 1.3.12.
-	 * A trailing slash will make Apache call add_cgi_vars 
-	 * recursively ad infinitum in some handlers. May be it's an Apache bug?
-	 */
 	if ((req->path_info) && 
-	     (req->path_info[strlen(req->path_info) - 1] == '/'))
+	    (req->path_info[strlen(req->path_info) - 1] == '/'))
 	{
 	    int i;
-	    i = strlen( req->path_info );
+	    i = strlen(req->path_info);
 	    /* take out the slash */
-	    req->path_info[ i - 1 ] = 0;
+	    req->path_info[i - 1] = 0;
 
 	    Py_BEGIN_ALLOW_THREADS;
 	    ap_add_cgi_vars(req);
 	    Py_END_ALLOW_THREADS;
-	    *request_obj = make_requestobject(req);
+	    request_obj = make_requestobject(req);
 
 	    /* put the slash back in */
-	    req->path_info[ i - 1 ] = '/'; 
-	    req->path_info[ i ] = 0;
+	    req->path_info[i - 1] = '/'; 
+	    req->path_info[i] = 0;
 	} 
 	else 
 	{ 
 	    Py_BEGIN_ALLOW_THREADS;
 	    ap_add_cgi_vars(req);
 	    Py_END_ALLOW_THREADS;
-	    *request_obj = make_requestobject(req);
+	    request_obj = make_requestobject(req);
 	}
-
+	
 	/* store the pointer to this object in notes */
-	sprintf(s2, "%d", (int) *request_obj);
+	/* XXX this is not good... */
+	sprintf(s2, "%d", (int) request_obj);
 	ap_table_set(req->notes, "python_request_ptr", s2);
+	return request_obj;
     }
 }
 
@@ -2104,24 +1855,24 @@ static int python_handler(request_rec *req, char *handler)
 #endif
 
     /* get/create obcallback */
-    obcallback = get_obcallback(interpreter, req->server);
+    obcallback = get_obcallback(interpreter, req);
 
     /* we must have a callback object to succeed! */
     if (!obcallback) 
     {
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req->server,
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req,
 		     "python_handler: get_obcallback returned no obCallBack!");
 	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
     /* find __interpreter__ in obCallBack */
-    tstate = (PyThreadState *) PyInt_AsLong(PyObject_GetAttrString(obcallback, INTERP_ATTR));
+    tstate = (PyThreadState *) PyCObject_AsVoidPtr(PyObject_GetAttrString(obcallback, INTERP_ATTR));
 
     /* make this thread state current */
     PyThreadState_Swap(tstate);
 
     /* create/acquire request object */
-    get_request_object(req, &request_obj);
+    request_obj = get_request_object(req, tstate);
 
     /* 
      * The current directory will be that of the last encountered 
@@ -2145,7 +1896,7 @@ static int python_handler(request_rec *req, char *handler)
 #endif
 
     if (! resultobject) {
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req->server, 
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req, 
 		     "python_handler: Dispatch() returned nothing.");
 	return HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -2153,7 +1904,7 @@ static int python_handler(request_rec *req, char *handler)
 	/* Attempt to analyse the result as a string indicating which
 	   result to return */
 	if (! PyInt_Check(resultobject)) {
-	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req->server, 
+	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, req, 
 			 "python_handler: Dispatch() returned non-integer.");
 	    return HTTP_INTERNAL_SERVER_ERROR;
 	}
